from __future__ import annotations
from typing import Iterator, Optional, Tuple, Union

import collections
import contextlib
import math
import numpy as np
import torch
import luz

__all__ = ["Score", "Scorer", "CrossValidationScorer", "HoldoutValidationScorer"]

Score = collections.namedtuple("Score", ["predictor", "score"])


class Scorer:
    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Union[torch.device, str],
    ) -> luz.Score:
        raise NotImplementedError


class CrossValidationScorer(Scorer):
    def __init__(self, num_folds: int, fold_seed: Optional[int] = None) -> None:
        self.num_folds = num_folds
        self.fold_seed = fold_seed

    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Union[torch.device, str],
    ) -> luz.Score:
        test_losses = []
        for fit_dataset, test_dataset in self._split_dataset(dataset):
            with luz.temporary_seed(
                fold_seed
            ) if self.fold_seed is not None else contextlib.nullcontext():
                _, score = learner.learn(
                    dataset=fit_dataset, test_dataset=test_dataset, device=device
                )

            test_losses.append(score)

        score = sum(test_losses) / self.num_folds

        return Score(learner.learn(dataset=dataset, device=device).predictor, score)

    def _split_dataset(
        self, dataset: luz.Dataset
    ) -> Iterator[Tuple[luz.Dataset, luz.Dataset]]:
        points_per_fold = math.ceil(len(dataset) // self.num_folds)
        fold_lengths = [points_per_fold] * self.num_folds
        fold_lengths[-1] -= sum(fold_lengths) - len(dataset)

        folds = dataset.random_split(lengths=fold_lengths)

        for i in range(self.num_folds):
            train_dataset = luz.ConcatDataset(
                [f for j, f in enumerate(folds) if j != i]
            )
            val_dataset = folds[i]

            yield train_dataset, val_dataset


class HoldoutValidationScorer(Scorer):
    def __init__(self, holdout_fraction: float) -> None:
        self.holdout_fraction = holdout_fraction

    def score(
        self,
        learner: luz.Learner,
        dataset: luz.Dataset,
        device: Union[torch.device, str],
    ) -> luz.Score:
        fit_dataset, test_dataset = self._split_dataset(dataset=dataset)

        return learner.learn(
            dataset=fit_dataset, test_dataset=test_dataset, device=device
        )

    def _split_dataset(self, dataset: luz.Dataset) -> Tuple[luz.Dataset, luz.Dataset]:
        l = np.random.permutation(len(dataset))
        ind = round(float(f"{self.holdout_fraction}e{luz.int_length(len(l))-1}"))
        holdout, remainder = np.split(l, [ind])

        fit_dataset = dataset.subset(remainder)
        test_dataset = dataset.subset(holdout)

        return fit_dataset, test_dataset



from luz.datasets import *
from luz.events import *
from luz.flags import *
from luz.handlers import *
from luz.learners import *
from luz.modules import *
from luz.optimizer import *
from luz.predictors import *
from luz.scorers import *
from luz.trainers import *
from luz.transforms import *
from luz.tuners import *
from luz.utils import *

# ------------------------------------------- #

# import luz

# def build_learner(config):
#     """Returns Python class reference for learning algorithm specified by config file.

#     Returns
#     -------
#     class
#         Python class for learning algorithm specified by config file.
#     """
#     cuda = config.nested_dictionary.get('cuda',False)# if 'cuda' in config else False
#     seed = config.nested_dictionary.get('seed')# if 'seed' in config else None

#     dataset = _recursive_build(nested_dictionary=config.nested_dictionary,keys=('dataset',))
#     scorer = _recursive_build(nested_dictionary=config.nested_dictionary,keys=('scorer',))
#     # FIXME: this is a clumsy location for assigning self.tuner
#     # FIXME: self.tuner is an issue - should be config.tuner
#     tuner = _recursive_build(nested_dictionary=config.nested_dictionary,keys=('tuner',))
#     config.tuner = tuner

#     return luz.Learner(config=config,dataset=dataset,scorer=scorer,cuda=cuda,seed=seed)

# def build_model_trainer(config, dataset=None, tuner=None):
#     """Returns Python class reference for learning algorithm specified by config file.

#     Returns
#     -------
#     class
#         Python class for learning algorithm specified by config file.
#     """
#     # FIXME: why does dataset have None as a default value? why shouldn't it be mandatory with no default?
#     if hasattr(config,'tuner') and config.tuner is not None:
#         tuned_config = config.tuner.tune(config=config,tuning_dataset=dataset)

#         model = _recursive_build(config=tuned_config, keys=('model',))
#         trainer = _recursive_build(config=tuned_config, keys=('trainer',))
#     else:
#         model = _recursive_build(nested_dictionary=config.nested_dictionary, keys=('model',))
#         trainer = _recursive_build(nested_dictionary=config.nested_dictionary, keys=('trainer',))

#     return model, trainer

# def _recursive_build(nested_dictionary, keys):
#     class_str, kwargs = next(iter(nested_dictionary.get(keys).d.items()))

#     if kwargs is None:
#         kwargs = {}
#     else:
#         for k,v in kwargs.items():
#             if hasattr(v,'items'):
#                 kwargs[k] = _recursive_build(nested_dictionary=nested_dictionary,keys=keys+(class_str,k))
#     return luz.utils.string_to_class(class_str=class_str)(**kwargs)
#     #return luz.utils.string_to_class(class_str=class_str)(*args,**kwargs)

# # def _recursive_build(config, keys):
# #     try:
# #         class_str, d = next(iter(config.config.get(keys).d.items()))
# #         print('class_str')
# #         print(class_str)
# #         print(d)
# #     except AttributeError: # k not in self.config
# #         return None
# #
# #     #FIXME: more elegant way of handling class strings with no args?
# #     #FIXME: is this just d??
# #     #kwargs = dict((k,_recursive_build(config=config,keys=keys+(class_str,k))) if hasattr(v,'items') else (k[0],v) for k,v in d.items())
# #     #print(kwargs)
# #     kwargs = {k: _recursive_build(config=config,keys=keys+(class_str,k)) if hasattr(v,'items') else v for k,v in d.items()} if hasattr(d,'items') else {}
# #     print(kwargs)
# #     return luz.utils.string_to_class(class_str=class_str)(**kwargs)

# ---------------------------------------------------- #

# import copy
# import os
# import uuid

# from ruamel.yaml import YAML

# import luz


# class Config:
#     """Object for reading and interpreting model configuration file.

#     Attributes
#     ----------
#     config_path: str
#         absolute path to configuration (YAML) file
#     """

#     def __init__(self, nested_dictionary):
#         self.nested_dictionary = nested_dictionary

#     @property
#     def hyperparameters(self):
#         hp_locations = tuple(
#             tuple(k) for *k, _ in self.nested_dictionary.find_key(k="$tune")
#         )
#         return tuple(
#             luz.config.Hyperparameter(
#                 location=k, **self.nested_dictionary[k]["$tune"].d
#             )
#             for k in hp_locations
#         )
#         # return dict(zip(hp_keys,hp_values))

#     @property
#     def pins(self):
#         pin_locations = tuple(
#             tuple(k) for *k, _ in self.nested_dictionary.find_key(k="$pin")
#         )
#         return tuple(
#             luz.config.Pin(
#                 location=k, pin_equation_string=self.nested_dictionary[k]["$pin"]
#             )
#             for k in pin_locations
#         )
#         # return dict(zip(pin_keys,pin_values))

#     def add_samples(self, samples):
#         sample = copy.deepcopy(self.nested_dictionary)

#         pin_vals = tuple(
#             pin.evaluate(hyperparameters=self.hyperparameters, samples=samples)
#             for pin in self.pins
#         )

#         sample.update(
#             {hp.location: val for hp, val in zip(self.hyperparameters, samples)}
#         )
#         sample.update({pin.location: val for pin, val in zip(self.pins, pin_vals)})

#         # pin_values = {keys: pin.evaluate(hyperparameter_values=dict(((hp.label,samp) for hp,samp in zip(config.hyperparameters,hyperparameter_samples)))) for keys,pin in config.pins.items()}
#         # pin_values = {keys: pin.evaluate(hyperparameter_values={hyperparameters[k].label: hyperparameter_samples[k] for k in hyperparameters}) for keys,pin in config.pins.items()}

#         return luz.Config(nested_dictionary=sample)

#     @classmethod
#     def read(cls, config_path):
#         """Reads configuration YAML file into nested dictionary.

#         Returns
#         -------
#         dictionary
#             Nested dictionary corresponding to raw configuiration YAML file.

#         """
#         with open(os.path.realpath(os.path.expanduser(config_path)), "r") as f:
#             nested_dictionary = luz.utils.NestedDictionary(
#                 cls.commented_map_to_dictionary(map=YAML().load(f))
#             )

#         return luz.Config(nested_dictionary=nested_dictionary)

#     @classmethod
#     def commented_map_to_dictionary(cls, map):
#         """Converts a CommentedMap object to a nested dictionary.

#         Parameters
#         ----------
#         map : CommentedMap
#             CommentedMap object corresponding to raw configuration YAML file to be converted to a nested dictionary.

#         Returns
#         -------
#         dictionary
#             Nested dictionary corresponding to raw configuiration YAML file.

#         """
#         if hasattr(map, "keys"):
#             map_dict = dict(map)
#             return {
#                 k: cls.commented_map_to_dictionary(map=map_dict[k])
#                 for k in map_dict.keys()
#             }
#         else:
#             return map

# -------------------------------------------------- #

# def __setattr__(self, name, value):
#     if self.obj is None:
#         samples = self.tuner.sample(tuning_parameters=self.tuning_parameters)
#         self.obj = self.cls(*self.args,**self.kwargs,**samples)
#     setattr(self.obj,name,value)

# def _make_wrapper_methods(self):
#     #FIXME: are we missing magic methods that might depend on hyperparameters?
#     for k,v in inspect.getmembers(self.cls,predicate=inspect.isroutine):#self.cls.__dict__.items():
#         if hasattr(v,'__call__') and not k.startswith('__') and not k.endswith('__'):
#             setattr(self,k,self._get_sample(v))
#
# def _get_sample(self, f):
#     #FIXME: checking if self.obj is not None should happen in wrapped, not the other way around
#     if self.obj is not None:
#         return f
#     else:
#         def wrapped(*args,**kwargs):
#             samples = self.tuner.sample(tuning_parameters=self.tuning_parameters)
#             self.obj = self.cls(*self.args,**self.kwargs,**samples)
#             return getattr(self.obj,f.__name__)(*args,**kwargs)
#         return wrapped

# -------------------------------------------------- #

import ast

# import operator
# import re


# class Pin:
#     def __init__(self, location, pin_equation_string):
#         self.location = location
#         self.pin_equation_string = pin_equation_string

#     def evaluate(self, hyperparameters, samples):
#         label_sample_dict = self._label_sample_dict(
#             hyperparameters=hyperparameters, samples=samples
#         )
#         pin_expression = self._pin_expression(label_sample_dict=label_sample_dict)
#         # parse pin equation string into operators
#         node = ast.parse(pin_expression, mode="eval").body

#         # evaluate operators to obtain the pin value
#         return self._evaluate_operators(node=node)

#     def _label_sample_dict(self, hyperparameters, samples):
#         labels = (hp.label for hp in hyperparameters)
#         return dict(zip(labels, samples))

#     def _pin_expression(self, label_sample_dict):
#         # substitute hyperparameters (given as dictionary whose keys are hp labels and whose values are hp values) into pin equation string

#         pattern = "|".join(label_sample_dict)
#         replacement = lambda m: str(label_sample_dict[re.escape(m.group(0))])
#         # replacement = lambda m: str(hyperparameter_values[re.escape(m.group(0))])
#         pin_expression = re.sub(
#             pattern=pattern, repl=replacement, string=self.pin_equation_string
#         )

#         return pin_expression

#     # def _replace(self, match, labels, samples):
#     #     return str(samples[labels.index(re.escape(m.group(0)))])

#     def _evaluate_operators(self, node):
#         ops = {
#             ast.Add: operator.add,
#             ast.Sub: operator.sub,
#             ast.Mult: operator.mul,
#             ast.Div: operator.truediv,
#             ast.Pow: operator.pow,
#             ast.BitXor: operator.xor,
#             ast.USub: operator.neg,
#             ast.FloorDiv: operator.floordiv,
#         }

#         if isinstance(node, ast.Num):  # <number>
#             return node.n
#         elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
#             return ops[type(node.op)](
#                 self._evaluate_operators(node.left),
#                 self._evaluate_operators(node.right),
#             )
#         elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
#             return ops[type(node.op)](self._evaluate_operators(node.operand))
#         else:
#             raise TypeError(node)



from __future__ import annotations
from typing import Any, Iterable, Iterator, Optional, Union

import collections
import contextlib
import copy
import luz
import numpy as np
import re

__all__ = ["Tuner", "BayesianTuner", "GridTuner", "RandomSearchTuner"]


class Sample:
    def __init__(
        self, lower: Union[int, float], upper: Union[int, float], tuner: luz.Tuner
    ) -> None:
        self.lower = lower
        self.upper = upper
        self.tuner = tuner

    def __call__(self) -> Union[int, float]:
        return self.tuner.sample(self.lower, self.upper)


class Choose:
    def __init__(self, choices: Iterable[Any], tuner: luz.Tuner) -> None:
        self.choices = tuple(choices)
        self.tuner = tuner

    def __call__(self) -> Any:
        return self.tuner.choose(self.choices)


class Pin:
    def __init__(self, equation: str, tuner: luz.Tuner) -> None:
        self.equation = equation
        self.tuner = tuner

    def __call__(self) -> Any:
        # NOTE: this sorting should ensure that longer variable names are substituted first,
        # which should prevent partial variable name clobbering
        pattern = "|".join(sorted(self.tuner.values, key=len)[::-1])
        replacement = lambda m: str(self.tuner.values[re.escape(m.group(0))])
        expression = re.sub(pattern=pattern, repl=replacement, string=self.equation)

        # parse pin equation string into operators and evaluate operators to obtain the pin value
        return luz.evaluate_expression(expression)


class Conditional:
    def __init__(
        self, condition: str, if_true: Any, if_false: Any, tuner: luz.Tuner
    ) -> None:
        self.pin = Pin(condition, tuner)
        self.if_true = if_true
        self.if_false = if_false

    def __call__(self) -> Any:
        return self.if_true if self.pin() else self.if_false


class Tuner:
    def __init__(
        self, num_iterations: int, scorer: luz.Scorer, seed_loop: Optional[bool] = False
    ) -> None:
        self.num_iterations = num_iterations
        self.scorer = scorer
        self.seed_loop = seed_loop

        self.values = {}

        self.scores = collections.defaultdict(list)
        self.best_predictor = None

    def hp_sample(self, lower: Union[int, float], upper: Union[int, float]) -> Sample:
        return Sample(lower, upper, tuner=self)

    def hp_choose(self, *choices: Any) -> Choose:
        return Choose(choices=choices, tuner=self)

    def pin(self, equation: str) -> Pin:
        return Pin(equation, tuner=self)

    def conditional(self, condition: str, if_true: Any, if_false: Any) -> Conditional:
        return Conditional(condition, if_true, if_false, tuner=self)

    def tune(self, **kwargs: Union[Choose, Conditional, Pin, Sample]) -> Iterator[Any]:
        # FIXME: implement optional skipping if an old set of hyperparameters is sampled
        for i in range(self.num_iterations):
            # seed to ensure reproducibility in each iteration of the tuning loop
            with luz.temporary_seed(i) if self.seed_loop else contextlib.nullcontext():
                for k, v in kwargs.items():
                    self.values[k] = v()

                values = tuple(self.values.values())

                if len(values) == 1:
                    yield from values
                else:
                    yield values

    def score(self, learner: luz.Learner, dataset, device):
        predictor, score = self.scorer.score(learner, dataset, device)

        if self.best_predictor is None or score < self.best_score:
            self.best_predictor = copy.deepcopy(predictor)

        self.scores[tuple(sorted(copy.deepcopy(self.values).items()))].append(score)

        return score

    @property
    def best_hyperparameters(self):
        mean_scores = {k: sum(v) / len(v) for k, v in self.scores.items()}
        try:
            k = min(mean_scores, key=mean_scores.get)
            return dict(k)
        except ValueError:
            return None

    @property
    def best_score(self):
        mean_scores = {k: sum(v) / len(v) for k, v in self.scores.items()}
        try:
            k = min(mean_scores, key=mean_scores.get)
            return mean_scores[k]
        except ValueError:
            return None

    def sample(self, lower, upper):
        raise NotImplementedError

    def choose(self, choices):
        raise NotImplementedError

    def get_sample(self):
        raise NotImplementedError

    def sample(self, tuning_parameters):
        raise NotImplementedError


class BayesianTuner(Tuner):
    def sample(self, lower, upper):
        raise NotImplementedError

    def choose(self, choices):
        raise NotImplementedError


class GridTuner(Tuner):
    def sample(self, lower, upper):
        raise NotImplementedError

    def choose(self, choices):
        raise NotImplementedError


class RandomSearchTuner(Tuner):
    def sample(self, lower, upper):
        (datatype,) = set((type(lower), type(upper)))
        if datatype == int:
            return np.random.randint(low=lower, high=upper + 1)
        elif datatype == float:
            return np.random.uniform(low=lower, high=upper)
        else:
            return None

    def choose(self, choices):
        return np.random.choice(a=choices)



from __future__ import annotations
from typing import Any, Callable, Optional

import ast
import collections
import contextlib
import functools
import importlib
import math
import numpy as np
import operator
import pathlib
import torch


__all__ = [
    "evaluate_expression",
    "expand_path",
    "int_length",
    "memoize",
    "set_seed",
    "string_to_class",
    "temporary_seed",
]

# from https://stackoverflow.com/a/9558001

ops = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.truediv,
    ast.Pow: operator.pow,
    ast.BitXor: operator.xor,
    ast.USub: operator.neg,
    ast.FloorDiv: operator.floordiv,
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
}


def evaluate_expression(expression: str) -> Any:
    node = ast.parse(expression, mode="eval").body

    return _evaluate_operators(node)


def _evaluate_operators(node) -> Any:
    # FIXME: type annotate `node`
    # <number>
    if isinstance(node, ast.Num):
        return node.n
    # <left> <operator> <right>
    elif isinstance(node, ast.BinOp):
        return ops[type(node.op)](
            _evaluate_operators(node.left), _evaluate_operators(node.right),
        )
    # <operator> <operand> e.g., -1
    elif isinstance(node, ast.UnaryOp):
        return ops[type(node.op)](_evaluate_operators(node.operand))
    elif isinstance(node, ast.Compare):
        # from https://github.com/danthedeckie/simpleeval/blob/master/simpleeval.py
        right = _evaluate_operators(node.left)
        to_return = True
        for operation, comp in zip(node.ops, node.comparators):
            if not to_return:
                break
            left = right
            right = _evaluate_operators(comp)
            to_return = ops[type(operation)](left, right)
        return to_return
    else:
        raise TypeError(node)


def expand_path(path: str, dir: Optional[str] = None) -> str:
    p = pathlib.Path(path).expanduser()
    if dir is not None:
        p = pathlib.Path(dir).joinpath(p)
    return str(p.expanduser().resolve())


def int_length(n: int) -> int:
    # from https://stackoverflow.com/a/2189827
    if n > 0:
        return int(math.log10(n)) + 1
    elif n == 0:
        return 1
    else:
        # +1 if you don't count the '-'
        return int(math.log10(-n)) + 2


T = Callable[..., Any]


def memoize(func: T) -> T:
    func.cache = collections.OrderedDict()

    @functools.wraps(func)
    def memoized(*args, **kwargs):
        k = (args, frozenset(kwargs.items()))
        if k not in func.cache:
            func.cache[k] = func(*args, **kwargs)

        return func.cache[k]

    return memoized


def set_seed(seed: int) -> None:
    torch.manual_seed(seed)
    np.random.seed(seed)


def string_to_class(class_str: str) -> Any:
    """Converts a string specifying the name of a Python class to a reference to that class.

    Parameters
    ----------
    default_source
        The source (i.e. module and submodule names) which will be used if class_str contains no periods.
    class_str
        The string specifying the name of a Python class. If class_str contains periods, then it is assumed that
        class_str contains the full module and submodule structure, and default_source will be ignored.

    Returns
    -------
    class object
        The class object (i.e. the class which can be called to create an instance of said class)
        corresponding to the provided string.

    """
    try:
        split_name = class_str.rsplit(".", 1)
    except AttributeError:
        # means that class_str doesn't have rsplit function - probably was passed as None
        return None

    try:
        source_path, class_name = split_name
        return getattr(importlib.import_module(source_path), class_name)
    except ValueError:
        # couldn't unpack split_name, so len(split_name) < 2 - probably = 1, no '.' in class_str so assume builtin
        return getattr(importlib.import_module("builtins"), class_str)


@contextlib.contextmanager
def temporary_seed(seed: int) -> None:
    # adapted from https://stackoverflow.com/a/49557127
    np_state = np.random.get_state()
    torch_state = torch.random.get_rng_state()

    set_seed(seed)

    try:
        yield
    finally:
        np.random.set_state(np_state)
        torch.random.set_rng_state(torch_state)



from __future__ import annotations
from typing import Any, Iterable, Optional

import copy
import contextlib
import torch

__all__ = ["Predictor"]


class Predictor:
    """
    A Predictor is an object which takes objects from a domain set and predicts the corresponding values from a label set.

    Attributes:
        model (torch.nn.Module): PyTorch module used for prediction.
    """

    def __init__(self, model: torch.nn.Module) -> None:
        self.model = model

    @classmethod
    def builder(cls, *args, **kwargs):
        def f():
            return cls(*args, **kwargs)

        return f

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.model.forward(x)

    def predict(self, x: luz.Data) -> torch.Tensor:
        with torch.no_grad(), self.eval():
            return self.__call__(x)

    def to(self, *args, **kwargs) -> luz.Predictor:
        self.model.to(*args, **kwargs)

        return self

    def train(self, mode: Optional[bool] = True) -> luz.Predictor:
        self.model.train(mode)

        return self

    @contextlib.contextmanager
    def eval(self) -> None:
        training = copy.copy(self.model.training)

        try:
            if training:
                self.model.eval()
            yield
        finally:
            if training:
                self.model.train()



from __future__ import annotations
from typing import Any, Callable, Optional, Type

import luz

__all__ = ["Learner"]


class Learner:
    """
    A Learner is an object which takes a dataset as input and produces a predictor as output. Learners are distinguished by the different protocols they use to construct a predictor from a given dataset.
    """

    def __init__(
        self, trainer: luz.Trainer, cls: Type[torch.nn.Module], *args, **kwargs,
    ) -> None:
        self.cls = cls
        self.trainer = trainer
        self.args = args
        self.kwargs = kwargs

    def learn(
        self,
        dataset: luz.Dataset,
        device: Union[str, torch.device],
        val_dataset: Optional[luz.Dataset] = None,
        test_dataset: Optional[luz.Dataset] = None,
    ) -> luz.Score:
        """
        Learn a predictor based on a given dataset.

        Args:
            dataset: Dataset which will be used to learn a predictor.

        Returns:
            luz.Predictor: Predictor which was learned using `dataset`.
        """
        p = luz.Predictor(self.cls(*self.args, **self.kwargs))

        self.trainer.run(
            predictor=p,
            dataset=dataset,
            val_dataset=val_dataset,
            device=device,
            train=True,
        )

        if test_dataset is None:
            return luz.Score(p, None)
        else:
            score = self.trainer.run(
                predictor=p, dataset=test_dataset, device=device, train=False
            )
            return luz.Score(p, score)



"""

Custom PyTorch modules. Each class must implement torch.nn.Module, and therefore must have an __init__ method and a forward method.

"""
from __future__ import annotations
from typing import Iterable, Optional, Tuple

import collections
import networkx as nx
import torch

__all__ = ["ElmanRNN", "FC", "FCRNN", "Module", "Reshape", "WAVE"]


class Module(torch.nn.Module):
    def __init__(self, module: torch.nn.Module, activation=None) -> None:
        super().__init__()
        self.module = module
        self.activation = activation or (lambda x: x)

    def forward(self, *args: torch.Tensor, **kwargs: torch.Tensor) -> torch.Tensor:
        return self.activation(self.module(*args, **kwargs))


class ElmanRNN(torch.nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=None,
        nonlinearity=None,
        bias=None,
        batch_first=None,
        dropout=None,
        bidirectional=None,
        h0=None,
    ):
        super().__init__()

        self.hidden_size = hidden_size

        kwargs = {}
        if num_layers:
            kwargs["num_layers"] = num_layers
        if nonlinearity:
            kwargs["nonlinearity"] = nonlinearity
        if bias:
            kwargs["bias"] = bias
        if batch_first:
            kwargs["batch_first"] = batch_first
        if dropout:
            kwargs["dropout"] = dropout
        if bidirectional:
            kwargs["bidirectional"] = bidirectional

        self.h0 = h0
        self.rnn = torch.nn.RNN(
            input_size=input_size, hidden_size=hidden_size, **kwargs
        )

    def forward(self, x):
        if self.h0:
            output, hidden = self.rnn.forward(input=x, h0=self.h0)
        else:
            output, hidden = self.rnn.forward(input=x)
        return output


class FC(torch.nn.Module):
    def __init__(self, in_features, out_features, *hidden_features):
        super().__init__()
        sizes = tuple((in_features, *hidden_features, out_features))
        layers = (
            torch.nn.Linear(in_features=n_in, out_features=n_out)
            for n_in, n_out in zip(sizes, sizes[1:])
        )
        self.seq = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.seq(x)


class FCRNN(torch.nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super().__init__()

        self.hidden_size = hidden_size
        self.hidden = self._init_hidden()

        self.i2h = torch.nn.Linear(input_size + hidden_size, hidden_size)
        self.i2o = torch.nn.Linear(input_size + hidden_size, output_size)

    def forward(self, x):
        self.hidden = self._init_hidden()

        for i in x.transpose(0, 1):
            combined = torch.cat((i, self.hidden), 1)
            self.hidden = self.i2h(combined)
            output = self.i2o(combined)

        return output

    def _init_hidden(self):
        return torch.zeros(1, self.hidden_size)


class Reshape(torch.nn.Module):
    def __init__(self, out_shape):
        super().__init__()
        self.shape = tuple(out_shape)

    def forward(self, x):
        return x.view(self.shape)


class WAVE(torch.nn.Module):
    def __init__(self, atom_feature_size: int, num_passes: int) -> None:
        super().__init__()

        self.d = atom_feature_size
        self.num_passes = num_passes

        self.T_to_A = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)
        self.C_to_A = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)
        self.T_to_B = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)
        self.C_to_B = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)
        self.W = torch.randn(size=(self.d, 1))
        self.W_u = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)
        self.W_o = torch.nn.Linear(in_features=2 * self.d, out_features=self.d)

        self.sigmoid = torch.nn.Sigmoid()
        self.relu = torch.nn.ReLU()

    def forward(self, x: torch.Tensor, edge_index: torch.Tensor) -> torch.Tensor:
        traversal_order = self._traverse_graph(x, edge_index)

        lengths = tuple(len(l) for l in traversal_order)
        perm = torch.eye(sum(lengths))[[a for b in traversal_order for a in b]]

        x = torch.matmul(perm, x).t()

        for _ in range(self.num_passes):
            self._propagate(x, lengths)
            self._propagate(x, lengths[::-1])

        return x

    def _traverse_graph(self, x: torch.Tensor, edge_index: torch.Tensor) -> Tuple[int]:
        g = nx.Graph(list(edge_index.t().numpy()))

        root, *_ = nx.center(g)

        digraph = nx.bfs_tree(g, root)

        order = [[root]]
        while sum(len(a) for a in order) < len(g):
            # FIXME: not sure what this ordered dict stuff actually does if anything...
            succs = list(
                collections.OrderedDict.fromkeys(
                    [r for s in order[-1] for r in list(digraph.successors(s))]
                ).keys()
            )
            order.append(succs)

        return order

    def _propagate(self, x: torch.Tensor, lengths: Iterable[int]) -> None:
        offset = 0
        for t, c in zip(lengths, lengths[1:]):
            i = offset + t
            j = i + c

            T = x[:, offset:i]
            C = x[:, i:j]
            N = x[:, offset:j]

            for k in range(c):
                s = C.narrow(dim=1, start=k, length=1)

                m = self._mix_gate(state=s, tree=T, cross=C, neighbors=N)

                u = self.sigmoid(self.W_u(torch.cat(tensors=[m, s], dim=0).t()).t())
                o = self.relu(self.W_o(torch.cat(tensors=[m, s], dim=0).t()).t())

                C[:, k : k + 1] = u * o + (1 - u) * m

            offset += t

    def _mix_gate(
        self,
        state: torch.Tensor,
        tree: torch.Tensor,
        cross: torch.Tensor,
        neighbors: torch.Tensor,
    ) -> torch.Tensor:
        input_T = torch.cat([tree, state.expand_as(tree)], dim=0).t()
        input_C = torch.cat([cross, state.expand_as(cross)], dim=0).t()

        A = torch.nn.functional.softmax(
            torch.cat(
                tensors=[self.T_to_A(input_T).t(), self.C_to_A(input_C).t()], dim=1,
            ),
            dim=1,
        ) * self.W.expand_as(neighbors)

        B = torch.nn.functional.softsign(
            torch.cat([self.T_to_B(input_T).t(), self.C_to_B(input_C).t()], dim=1,)
        )

        return torch.sum((A + B) * neighbors, dim=1).unsqueeze(1)



from __future__ import annotations
from typing import Any, Callable, Type

import luz

__all__ = ["Optimizer"]


class Optimizer:
    def __init__(
        self, optim_cls: Type[torch.optim.Optimizer], *args: Any, **kwargs: Any
    ) -> None:
        self.optim_cls = optim_cls
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def builder(*args: Any, **kwargs: Any) -> Callable[..., luz.Optimizer]:
        def f(*new_args: Any, **new_kwargs: Any) -> luz.Optimizer:
            return luz.Optimizer(*args, *new_args, **kwargs, **new_kwargs)

        return f

    def link(self, predictor: luz.Predictor) -> torch.optim.Optimizer:
        return self.optim_cls(
            params=predictor.model.parameters(), *self.args, **self.kwargs
        )



from __future__ import annotations
from typing import Any, Iterable, Iterator, Optional, Tuple, Union

import itertools
import luz
import math
import pathlib
import torch

__all__ = [
    "Data",
    "Dataset",
    # "ChainDataset",
    "ConcatDataset",
    # "IterableDataset",
    "OnDiskDataset",
    "Subset",
    "UnpackDataset",
    "WrapperDataset",
]


class Data:
    def __init__(
        self, x: torch.Tensor, y: torch.Tensor = None, **kwargs: torch.Tensor
    ) -> None:
        self.d = {"x": x, "y": y, **kwargs}

        for k, v in dict(x=x, y=y, **kwargs).items():
            setattr(self, k, v)

    @property
    def keys(self):
        return self.d.keys()

    def __getitem__(self, k: str) -> Any:
        return self.d[k]

    def to(self, device: Union[str, torch.device]) -> luz.Data:
        for k, v in self.d.items():
            if torch.is_tensor(v):
                self.d[k] = v.to(device)

        return self

    def __repr__(self) -> str:
        kw_str = ",".join(f"{k}={v}" for k, v in self.d.items())
        return f"Data({kw_str})"


class BaseDataset:
    def _collate(self, batch: Iterable[luz.Data]) -> luz.Data:
        b, *_ = batch
        kw = {
            k: torch.stack([torch.as_tensor(sample[k]) for sample in batch], dim=0)
            for k in b.keys
        }

        return Data(**kw)

    def loader(
        self,
        batch_size: Optional[int] = 1,
        shuffle: Optional[bool] = True,
        num_workers: Optional[int] = 1,
        pin_memory: Optional[bool] = True,
        transform: Optional[luz.Transform] = None,
    ) -> torch.utils.data.Dataloader:

        if transform is None:
            collate_fn = self._collate
        else:
            collate_fn = lambda batch: transform(self._collate(batch))

        return torch.utils.data.DataLoader(
            dataset=self,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
            collate_fn=collate_fn,
        )

    def __add__(self, other: luz.Dataset) -> luz.ConcatDataset:
        return ConcatDataset([self, other])

    def subset(self, indices):
        return Subset(dataset=self, indices=indices)

    def random_split(self, lengths: Iterable[int]) -> Tuple[BaseDataset]:
        # adapted from https://pytorch.org/docs/stable/_modules/torch/utils/data/dataset.html#random_split
        indices = torch.randperm(sum(lengths)).tolist()
        return tuple(
            Subset(self, indices[offset - l : offset])
            for offset, l in zip(itertools.accumulate(lengths), lengths)
        )


class Dataset(BaseDataset, torch.utils.data.Dataset):
    """
    A Dataset is an object which contains, or at least can systematically access, points from some domain and optionally their associated labels.
    """

    def __init__(self, data: Iterable[luz.Data]) -> None:
        self.data = tuple(data)

    def __getitem__(self, index: int) -> luz.Data:
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)


class ConcatDataset(BaseDataset, torch.utils.data.ConcatDataset):
    pass


class Subset(BaseDataset, torch.utils.data.Subset):
    pass


class OnDiskDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, root: str) -> None:
        self.root = luz.expand_path(root)

    def __getitem__(self, index: int) -> luz.Data:
        return torch.load(pathlib.Path(self.root, f"{index}.pt"))

    def __len__(self) -> int:
        return len(tuple(pathlib.Path(self.root).glob("[0-9]*.pt")))


# class TensorDataset(BaseDataset,torch.utils.data.TensorDataset):
#     def __init__(self, **tensors: torch.Tensor) -> None:
#         self.tensors = tensors

#     def __getitem__(self, index: int) -> luz.Data:
#         return


class UnpackDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, keys, dataset) -> None:
        self.keys = keys
        self.dataset = dataset

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int) -> luz.Data:
        return Data(**dict(zip(self.keys, self.dataset[index])))


class WrapperDataset(BaseDataset, torch.utils.data.Dataset):
    def __init__(self, **datasets: torch.utils.data.Dataset) -> None:
        self.datasets = datasets

    def __len__(self) -> int:
        d, *_ = self.datasets.values()
        return len(d)

    def __getitem__(self, index: int) -> Any:
        return Data(**{k: d[index] for k, d in self.datasets.items()})


# class IterableDataset(torch.utils.data.IterableDataset):
#     def __init__(
#         self, dataset: Optional[torch.utils.data.IterableDataset] = None, **iterables: Iterable[torch.Tensor],
#     ) -> None:
#         if dataset is not None:
#             self.dataset = dataset
#         else:
#             iterable = (Data(**{k:v for k,v in zip(iterables.keys(),tensors)}) for tensors in zip(*iterables.values()))
#             self.dataset = _IterableDataset(iterable)

#     def _collate(self, batch) -> luz.Data:
#         b, *_ = batch
#         kw = {k: torch.stack([sample[k] for sample in batch], 0) for k in b.keys}

#         return Data(**kw)

#     def __len__(self) -> int:
#         return len(self.dataset)

#     def __add__(self, other: luz.IterableDataset) -> luz.IterableDataset:
#         return ChainDataset([self, other])

#     def __iter__(self) -> Iterator[Data]:
#         yield from self.dataset

#     def subset(self, indices):
#         return type(self)(dataset=Subset(dataset=self.dataset, indices=indices))

#     def loader(
#         self,
#         batch_size: Optional[int] = 1,
#         num_workers: Optional[int] = 1,
#         pin_memory: Optional[bool] = True,
#         transform: Optional[luz.Transform] = None,
#     ) -> torch.utils.data.Dataloader:

#         return torch.utils.data.DataLoader(
#             dataset=self.dataset,
#             batch_size=batch_size,
#             num_workers=num_workers,
#             pin_memory=pin_memory,
#             collate_fn=self._collate if transform is None else lambda batch: transform(self._collate(batch)),
#         )


# class _IterableDataset(torch.utils.data.IterableDataset):
#     def __init__(self, iterable: Iterable[luz.Data]) -> None:
#         self.iterable = tuple(iterable)

#     def __iter__(self) -> Iterator[Any]:
#         yield from self.iterable

#     def __len__(self) -> int:
#         return len(self.iterable)


class ChainDataset(BaseDataset, torch.utils.data.ChainDataset):
    pass



from __future__ import annotations
from typing import Any, Dict, Iterable, Optional, Tuple, Union

__all__ = ["Trainer", "SupervisedTrainer"]

import luz
import torch


class Trainer:
    def __init__(
        self,
        loss=None,
        optimizer: Optional[luz.Optimizer] = None,
        start_epoch: Optional[int] = 1,
        stop_epoch: Optional[int] = 2,
        handlers: Optional[Iterable[luz.Handler]] = None,
        process_batch: Optional[
            Callable[luz.Data, Tuple[torch.Tensor, Optional[torch.Tensor]]]
        ] = None,
        **loader_kwargs: Union[int, bool, luz.Transform],
    ) -> None:
        self.loss = loss
        self.optimizer = optimizer
        self.start_epoch = start_epoch
        self.stop_epoch = stop_epoch
        self.handlers = tuple(handlers or [])
        self.process_batch = process_batch or (lambda batch: (batch.x, batch.y))
        self.loader_kwargs = loader_kwargs

        self.flag = None
        self.state = {}

    def _call_event(self, event):
        for handler in self.handlers:
            getattr(handler, event.name.lower())(**self.state)

    def process_batch(
        self, batch: luz.Data
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        raise NotImplementedError

    def run_batch(
        self,
        predictor: luz.Predictor,
        data: torch.Tensor,
        target: torch.Tensor,
        device: Union[str, torch.device],
        optimizer: Optional[torch.optim.Optimizer] = None,
    ) -> float:
        raise NotImplementedError

    def run(
        self,
        predictor: luz.Predictor,
        dataset: luz.Dataset,
        device: Union[str, torch.device],
        train: bool,
        val_dataset: Optional[luz.Dataset] = None,
    ) -> None:
        self.set_mode(predictor, train)
        self.migrate(predictor, device)

        if train:
            # NOTE: must come after model.to
            optimizer = self.optimizer.link(predictor=predictor)

        loader = dataset.loader(**self.loader_kwargs)

        if train:
            self.state = dict(
                flag=luz.Flag.TRAINING,
                trainer=self,
                predictor=predictor,
                optimizer=optimizer,
                loader=loader,
            )
            self._call_event(event=luz.Event.TRAINING_STARTED)
        else:
            self.state = dict(
                flag=luz.Flag.TESTING, trainer=self, predictor=predictor, loader=loader,
            )
            self._call_event(event=luz.Event.TESTING_STARTED)

        if train:
            start, stop = self.start_epoch, self.stop_epoch
        else:
            start, stop = 1, 2

        for epoch in range(start, stop):
            running_loss = 0.0

            self.state.update(epoch=epoch)
            self._call_event(event=luz.Event.EPOCH_STARTED)

            for i, batch in enumerate(loader):
                data, target = self.process_batch(batch)
                self.state.update(ind=i, data=data, target=target)
                self._call_event(event=luz.Event.BATCH_STARTED)

                # migrate the input and target tensors to the appropriate device
                data, target = data.to(device), target.to(device)

                if train:
                    self.run_batch(predictor, data, target, device, optimizer)
                    if val_dataset is not None:
                        # FIXME: implement this!
                        raise NotImplementedError
                else:
                    running_loss += self.run_batch(predictor, data, target, device)
                # from https://coolnesss.github.io/2019-02-05/pytorch-gotchas
                # All of the variables defined above are now out of scope!
                # On CPU, they are already deallocated. On GPU, they will be deallocated soon.

                # Make sure deallocation has taken place
                if torch.cuda.is_available():
                    torch.cuda.synchronize()

                self._call_event(event=luz.Event.BATCH_ENDED)
            self._call_event(event=luz.Event.EPOCH_ENDED)

        if train:
            self._call_event(event=luz.Event.TRAINING_ENDED)
        else:
            self._call_event(event=luz.Event.TESTING_ENDED)

            return running_loss / len(loader)

    def set_mode(self, predictor: luz.Predictor, train: bool) -> None:
        predictor.train() if train else predictor.eval()

    def migrate(
        self, predictor: luz.Predictor, device: Union[str, torch.device]
    ) -> None:
        predictor.to(device=device)

    def backward(self, loss: torch.Tensor) -> None:
        loss.backward()

    def optimizer_step(self, optimizer: torch.optim.Optimizer) -> None:
        optimizer.step()
        optimizer.zero_grad()


class SupervisedTrainer(Trainer):
    def run_batch(
        self,
        predictor: luz.Predictor,
        data: torch.Tensor,
        target: torch.Tensor,
        device: Union[str, torch.device],
        optimizer: Optional[torch.optim.Optimizer] = None,
    ) -> float:
        output = predictor(data)
        loss = self.loss(output, target)

        if optimizer is not None:
            self.backward(loss)
            self.optimizer_step(optimizer)

        self.state.update(output=output, loss=loss)

        return loss.item()


class SemisupervisedTrainer(Trainer):
    pass


class UnupervisedTrainer(Trainer):
    pass



"""

Contains callback objects which perform various function during the training process. Every callback
should inherit the Callback class.

"""

# FIXME: Write a test for every handler here (many are likely broken)
# FIXME: Type annotate every handler here

from __future__ import annotations
from typing import Optional, Set

import datetime
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.transforms as mtransforms
import luz
import numpy as np
import torch

__all__ = [
    "Handler",
    # "Accuracy",
    # "AVP",
    # "Checkpoint",
    # "DataAndFit",
    # "DurbinWatson",
    # "EarlyStopping",
    # "FBeta",
    # "LogToFile",
    "Loss",
    "Progress",
    # "RVP",
    "Timer",
]


class Handler:
    def __init__(self):
        pass

    def batch_started(self, **kwargs):
        pass

    def batch_ended(self, **kwargs):
        pass

    def epoch_started(self, **kwargs):
        pass

    def epoch_ended(self, **kwargs):
        pass

    def testing_started(self, **kwargs):
        pass

    def testing_ended(self, **kwargs):
        pass

    def training_started(self, **kwargs):
        pass

    def training_ended(self, **kwargs):
        pass


class Accuracy(Handler):
    def __init__(self, y_transform=None, output_transform=None):
        # FIXME: test this class's performance
        self.y_transform = y_transform
        self.output_transform = output_transform

    def training_started(self, **kwargs):
        self.num_correct = 0
        self.num_points = 0

    def batch_ended(self, target, output, **kwargs):
        # FIXME: is this sufficiently general?
        # self.num_correct += torch.sum(y == torch.argmax(input=output,dim=-1))
        # self.num_points += y.shape[0]

        self.num_correct += torch.sum(
            (target if self.y_transform is None else self.y_transform.transform(target))
            == (
                output
                if self.output_transform is None
                else self.output_transform.transform(output)
            ).long()
        ).item()  # FIXME: this .long() is an ugly hardcoded error fix
        self.num_points += target.shape[
            0
        ]  # FIXME: assumes the first dimension is the batch size - is this always true?

    def training_ended(self, epoch, **kwargs):
        print(
            f"[Epoch {epoch}] Classification accuracy: {self.num_correct/self.num_points}"
        )


class AVP(Handler):
    def __init__(self):
        self.actual = []
        self.predicted = []

    def update(self, x, y, predicted):
        self.predicted.append(predicted.item())
        self.actual.append(y.item())

    def compute(self):
        self.fig, self.ax = plt.subplots()
        self.ax.scatter(self.actual, self.predicted, c="black")
        mmax = max([max(self.actual), max(self.predicted)])
        plt.xlim([0, mmax])
        plt.ylim([0, mmax])
        line = mlines.Line2D([0, 1], [0, 1], color="red")
        line.set_transform(self.ax.transAxes)
        self.ax.add_line(line)
        plt.ylabel("Predicted")
        plt.xlabel("Actual")
        plt.title("Predicted vs. actual")
        plt.show()


class Checkpoint(Handler):
    def __init__(self, save_interval, model_name, save_dir=None):
        if save_interval < 0:
            raise ValueError(
                "The number of epochs between checkpoints must be a positive integer."
            )

        self.model_name = model_name
        self.save_dir = luz.utils.expand_path(path=save_dir)

        # try:
        #     os.makedirs(self.save_dir)
        # except FileExistsError:
        #     pass
        try:
            os.makedirs(self.save_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        self.save_interval = save_interval

    # # FIXME: Fix this
    # def state_dict(self):
    #     return {}

    def epoch_ended(self, model, epoch, **kwargs):
        if epoch % self.save_interval == 0:
            save_path = os.path.join(
                self.save_dir, f"{model.model_name}_{epoch}.pth.tar"
            )
            torch.save(obj=model.state_dict(), f=save_path)


class DataAndFit(Handler):
    def __init__(self):
        self.xs = []
        self.actual = []
        self.predicted = []

    def update(self, x, y, predicted):
        self.xs.append(x.item())
        self.actual.append(y.item())
        self.predicted.append(predicted.item())

    def compute(self):
        plt.scatter(self.xs, self.actual, label="Actual data")
        plt.scatter(self.xs, self.predicted, label="Fit function")
        plt.xlabel("Input")
        plt.ylabel("Output")
        plt.title("Data and fit")
        plt.show()


class DurbinWatson(Handler):
    def __init__(self):
        self.numerator = 0
        self.denominator = 0
        self.last_residual = None

    def batch_ended(self, x, y, output, **kwargs):
        residual = (y - output).item()
        if self.last_residual is not None:
            self.numerator += (residual - self.last_residual) ** 2
        self.last_residual = residual
        self.denominator += residual ** 2

    def compute(self):
        print("DW: {0}".format(self.numerator / self.denominator))
        return self.numerator / self.denominator


class EarlyStopping(Handler):
    def __init__(self, monitor="val_loss", delta_thresh=0.0, patience=5):
        # FIXME: This might not work with the revised scorer scheme - investigate and repair/rewrite this whole class

        if delta_thresh < 0:
            raise ValueError("Threshold value must be nonnegative.")

        self.monitor = monitor
        self.delta_thresh = delta_thresh
        self.patience = patience
        self.wait = patience

        self.history = None

    def compile(self, model, trainer):
        if self.monitor == "train_loss":
            self.history = model.train_history
        elif self.monitor == "val_loss":
            self.history = model.val_history
        else:
            raise ValueError(
                "Monitor keyword {0} not recognized for EarlyStopping callback.".format(
                    self.monitor
                )
            )

    def on_epoch_end(self):
        if self.wait == 0:
            return False

        if (min(self.history) - self.history[-1]) < self.delta_thresh:
            self.wait -= 1
        else:
            self.wait = self.patience

        return True


class FBeta(Handler):
    def __init__(self, beta):
        self.beta = beta

        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def update(self, x, y, predicted):
        self.true_positive += y.item() == predicted.item()
        self.predicted_positive += predicted.item() == 1
        self.actual_positive += y.item() == 1

    def compute(self):
        precision = self.true_positive / self.predicted_positive
        recall = self.true_positive / self.actual_positive

        return (
            (1 + self.beta ** 2)
            * precision
            * recall
            / (precision * (self.beta ** 2) + recall)
        )


class LogToFile(Handler):
    def __init__(self, log_path, log_list):
        # FIXME: rewrite this whole class
        assert False, "TODO: Finish writing LogToFile"

        self.log_path = log_path
        self.log_list = log_list

    # def _compute_confusion_matrix(self):

    def on_epoch_end(self):
        with open(self.log_path, "w") as f:
            for log_compute in self.log_list:
                f.write(log_compute)


class Loss(Handler):
    def __init__(self, print_interval: Optional[float] = 0.25) -> None:
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.print_interval = print_interval

    def epoch_started(self, **kwargs: Any) -> None:
        self.running_loss = 0

    def batch_ended(
        self,
        flag: luz.Flag,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        loss: torch.Tensor,
        ind: int,
        **kwargs: Any,
    ) -> None:
        if flag == luz.Flag.TRAINING:
            # NOTE: it's very important to add loss.item() (as opposed to loss) to avoid a memory leak!
            self.running_loss += loss.item()  # *loader.batch_size
            total = len(loader.sampler) / loader.batch_size
            batch_interval = int(total * self.print_interval)
            cur = ind + 1
            if cur % batch_interval == 0 or cur == total:
                print(
                    f"[Epoch {epoch}] Average running loss: {self.running_loss / cur}."
                )


class Progress(Handler):
    def __init__(self, print_interval: Optional[float] = 0.25) -> None:
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.bar_length = 30
        self.print_interval = print_interval

    def batch_ended(
        self,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        flag: luz.Flag,
        ind: int,
        **kwargs: Any,
    ) -> None:
        if flag == luz.Flag.TRAINING:
            total = len(loader.sampler) / loader.batch_size
            batch_interval = max(int(total * self.print_interval), 1)
            cur = ind + 1
            if cur % batch_interval == 0 or cur == total:
                progress = cur / total
                num_eqs = int(progress * self.bar_length)
                num_spaces = self.bar_length - num_eqs
                bar = "=" * num_eqs + ">" + " " * num_spaces
                print(f"[Epoch {epoch}]: [{bar}] {int(100*progress)}%")

    def training_started(self, **kwargs: Any) -> None:
        print("Training started.")

    def training_ended(self, **kwargs: Any) -> None:
        print("Training ended.")

    def testing_started(self, flag, **kwargs: Any) -> None:
        print("Testing started.")

    def testing_ended(self, flag, **kwargs: Any) -> None:
        print("Testing complete.")


class RVP(Handler):
    def __init__(self):
        self.predicted = []
        self.residual = []

    def batch_ended(
        self, x: torch.Tensor, y: torch.Tensor, output: torch.Tensor, **kwargs: Any
    ) -> None:
        self.predicted.append(output.item())
        self.residual.append((y - output).item())

    def training_ended(self, **kwargs: Any) -> None:
        self.fig, self.ax = plt.subplots()
        self.ax.scatter(self.predicted, self.residual, c="black")
        mmax = max([abs(max(self.residual)), abs(min(self.residual))])
        plt.ylim([-mmax, mmax])
        plt.axhline(y=0.0, color="r", linestyle="-")
        plt.ylabel("Residual")
        plt.xlabel("Predicted")
        plt.title("Residual vs. predicted")
        plt.show()


class Timer(Handler):
    def __init__(self) -> None:
        self.start_time = 0
        self.end_time = 0

    def epoch_started(self, **kwargs: Any) -> None:
        self.start_time = datetime.datetime.now().replace(microsecond=0)

    def epoch_ended(self, epoch: int, **kwargs: Any) -> None:
        self.end_time = datetime.datetime.now().replace(microsecond=0)

        print(f"[Epoch {epoch}]: {self.end_time-self.start_time} to train")
